/*
 * Copyright 2007, Thomas Maier-Komor
 * Licenses: LGPL, see file lgpl.txt
 */

#ifndef _REENTRANT
#define _REENTRANT
#endif
#include <assert.h>
#include <dlfcn.h>
#include <errno.h>
#include <fcntl.h>
#include <limits.h>
#include <link.h>
#include <pthread.h>
#include <semaphore.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <strings.h>


typedef enum {
	cl /*close*/,
	fc /*fclose*/,
	wr /*write*/,
	fs /*fsync*/,
	ff /*fflush*/,
	tr /*terminate*/
} task_t;

typedef struct op
{
	task_t op;
	void *buf;
	size_t s,bs;
	struct op *next;
} op_t;

typedef struct operator
{
	int fd, wr_err;
	volatile size_t mem;
	char *fn;
	FILE *str;
	op_t *volatile head, *volatile tail, *volatile unused;
	pthread_t thread;
	pthread_mutex_t lock;
	pthread_cond_t oom;
	sem_t qlen;
	struct operator *next;
} operator_t;

static operator_t *Operators = 0;
static void *libc_hdl = 0;
static int Verbose = 0, Priority = 10, MaxThreads = 32, NumThreads = 0,
	   NumDirect = 0;
static size_t MemPerThread = 0;
static sem_t Idle;
static int (*realclose)(int) = 0,
	   (*realfsync)(int) = 0,
	   (*realfflush)(FILE *) = 0,
	   (*realwrite)(int, const void *, size_t) = 0,
	   (*realread)(int, void *, size_t) = 0,
	   (*realopen)(const char *, int, ...) = 0,
	   (*realopen64)(const char *, int, ...) = 0,
	   (*realopenat)(int, const char *, int, ...) = 0,
	   (*realopenat64)(int, const char *, int, ...) = 0,
	   (*realfclose)(FILE *) = 0,
	   (*realcreat)(const char *, mode_t) = 0,
	   (*realcreat64)(const char *, mode_t) = 0;
static FILE *
	   (*realfdopen)(int,const char *) = 0;
static void (*realexit)(int) = 0;
static void *writer_thread(void *);
static void error(const char *str, ...);
static void debug(const char *str, ...);
static void log(const char *str, ...);



void init(void)
{
	const char *V,*T,*P,*M;
	int err;
	if (libc_hdl != 0)
		return;
#ifdef __linux
	libc_hdl = dlopen("libc.so.6",RTLD_NOW|RTLD_LOCAL);
#else
	libc_hdl = dlopen("libc.so",RTLD_NOW|RTLD_LOCAL);
#endif
	assert(libc_hdl);
	realopen64 = (int(*)(const char *,int,...)) dlsym(libc_hdl,"open64");
	assert(realopen64);
	realopen = (int(*)(const char *,int,...)) dlsym(libc_hdl,"open");
	assert(realopen);
	realfdopen = (FILE *(*)(int,const char *)) dlsym(libc_hdl,"fdopen");
	assert(realfdopen);
	realfclose = (int (*)(FILE *)) dlsym(libc_hdl,"fclose");
	assert(realfclose);
	realcreat64 = (int(*)(const char *,mode_t)) dlsym(libc_hdl,"creat64");
	assert(realcreat64);
	realcreat = (int(*)(const char *,mode_t)) dlsym(libc_hdl,"creat");
	assert(realcreat);
#ifdef __sun
	realopenat64 = (int(*)(int,const char *,int,...)) dlsym(libc_hdl,"openat64");
	assert(realopenat64);
	realopenat = (int(*)(int,const char *,int,...)) dlsym(libc_hdl,"openat");
	assert(realopenat);
#endif
	realclose = (int(*)(int)) dlsym(libc_hdl,"close");
	assert(realclose);
	realfsync = (int(*)(int)) dlsym(libc_hdl,"fsync");
	assert(realfsync);
	realfflush = (int(*)(FILE *)) dlsym(libc_hdl,"fflush");
	assert(realfflush);
	realread = (int(*)(int, void *, size_t)) dlsym(libc_hdl,"read");
	assert(realread);
	realwrite = (int(*)(int, const void *, size_t)) dlsym(libc_hdl,"write");
	assert(realwrite);
	realexit = (void(*)(int)) dlsym(libc_hdl,"exit");
	assert(realexit);
	if (V = getenv("MTWRITE_VERBOSE"))
		Verbose = atol(V);
	if (T = getenv("MTWRITE_MAXTHREADS")) {
		MaxThreads = atol(T);
		log("Maximum number of wroker threads set to %d.\n",MaxThreads);
	}
	if (P = getenv("MTWRITE_PRIORITY"))
		Priority = atol(P);
	if (M = getenv("MTWRITE_MEMPERTHREAD")) {
		char dim,s;
		debug("MTWRITE_MEMPERTHREAD=\"%s\"\n",M);
		switch (s = sscanf(M,"%ld%c",&MemPerThread,&dim)) {
		default:
			error("Invalid MTWRITE_MEMPERTHREAD specification ignored.\n");
			break;
		case 2:
			switch (dim) {
			case 'K':
			case 'k':
				MemPerThread <<= 10;
				break;
			case 'M':
				MemPerThread <<= 20;
				break;
			case 'G':
				MemPerThread <<= 30;
				break;
			default:
				error("Invalid MTWRITE_MEMPERTHREAD specification ignored.\n");
				break;
			}
			/*FALLTHRU*/
		case 1:
			log("Maximum memory per thread: %ld\n",MemPerThread);
			break;
		}
	}
	err = sem_init(&Idle,0,MaxThreads ? MaxThreads : INT_MAX);
	assert(err != -1);
	if (Verbose)
		log(	"\nMT-write version r70317\n"
			"Copyright 2007, Thomas Maier-Komor\n"
			"Licenses: LGPL, see file lgpl.txt for details\n"
			"=============================================\n");
	debug("init successful (Verbose = %d,MaxThreads = %d, MemPerThread = %lld)\n",
			Verbose,MaxThreads,MemPerThread);
}


#ifdef __sun
#define init()
#endif


static void writelog(const char *m)
{
	static pthread_mutex_t lock = PTHREAD_MUTEX_INITIALIZER;
	int err;
	err = pthread_mutex_lock(&lock);
	assert(err == 0);
	(void) realwrite(2/*STDERR_FILENO*/,m,strlen(m));
	err = pthread_mutex_unlock(&lock);
	assert(err == 0);
}
	
static void error(const char *str, ...)
{
	va_list val;
	char msg[256] = "mtwrite.so: ";
	va_start(val,str);
	assert(libc_hdl != 0);
	(void) vsnprintf(msg+12,sizeof(msg)-12,str,val);
	writelog(msg);
	va_end(val);
}


static void log(const char *str, ...)
{
	if (Verbose >= 1) {
		char msg[256] = "mtwrite.so: ";
		va_list val;
		va_start(val,str);
		assert(libc_hdl != 0);
		(void) vsnprintf(msg+12,sizeof(msg)-12,str,val);
		writelog(msg);
		va_end(val);
	}
}


static void debug(const char *str, ...)
{
	if (Verbose >= 2) {
		va_list val;
		va_start(val,str);
		char msg[256] = "mtwrite.so: ";
		assert(libc_hdl != 0);
		(void) vsnprintf(msg+12,sizeof(msg)-12,str,val);
		writelog(msg);
		va_end(val);
	}
}


static void assignthread(int fd,const char *fn)
{
	int err;
	operator_t *op = Operators;
	pthread_attr_t attr;
	struct sched_param param;

	err = sem_wait(&Idle);
	assert(err == 0);
	while (op) {
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		assert(op->fd != fd);
		if ((op->head == 0) && (op->fd == -1)) {
			/* thread went out of work */
			int ql;
			op->fd = fd;
			op->fn = strdup(fn);
			op->str = 0;
			op->wr_err = 0;
			op->mem = 0;
			sem_getvalue(&op->qlen,&ql);
			assert(ql == 0);
			err = pthread_mutex_unlock(&op->lock);
			assert(err == 0);
			debug("assignthread: reusing thread %d\n",op->thread);
			return;
		}
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		op = op->next;
	}
	op = malloc(sizeof(operator_t));
	op->fd = fd;
	op->fn = strdup(fn);
	op->str = 0;
	op->wr_err = 0;
	op->mem = 0;
	err = pthread_mutex_init(&op->lock,0);
	assert(err == 0);
	err = pthread_cond_init(&op->oom,0);
	assert(err == 0);
	err = sem_init(&op->qlen,0,0);
	assert(err == 0);
	op->head = op->tail = op->unused = 0;
	(void) pthread_attr_init(&attr);
	(void) pthread_attr_getschedparam(&attr,&param);
	param.sched_priority = Priority;
	(void) pthread_attr_setschedparam(&attr,&param);
	err = pthread_create(&op->thread,&attr,writer_thread,op);
	if (err != 0) {
		error("cannot create new worker thread: %s\n",strerror(errno));
		free(op);
		return;
	}
	(void) pthread_attr_destroy(&attr);
	++NumThreads;
	debug("assignthread: created thread %d\n",op->thread);
	assert(err == 0);
	op->next = Operators;
	Operators = op;
}


int open64(const char *fn, int flags, ...)
{
	int ret;

	init();
	if (flags & O_CREAT) {
		va_list val;
		mode_t mode;

		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
		ret = realopen64(fn,flags,mode);
		debug("open64(\"%s\",%x,%x)= %d\n",fn,flags,mode,ret);
	} else {
		ret = realopen64(fn,flags);
		debug("open64(\"%s\",%x)= %d\n",fn,flags,ret);
	}
	if (((flags & (O_RDONLY|O_WRONLY|O_RDWR)) == O_WRONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}


int open(const char *fn, int flags, ...)
{
	int ret;

	init();
	if (flags & O_CREAT) {
		va_list val;
		mode_t mode;

		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
		ret = realopen(fn,flags,mode);
		debug("open(\"%s\",%x,%x) = %d\n",fn,flags,mode,ret);
	} else {
		ret = realopen(fn,flags);
		debug("open(\"%s\",%x) = %d\n",fn,flags,ret);
	}
	if (((flags & (O_RDONLY|O_WRONLY|O_RDWR)) == O_WRONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}


int creat(const char *path, mode_t mode)
{
	int ret;
	init();
	ret = realcreat(path,mode);
	debug("creat(\"%s\",%x) = %d\n",path,mode,ret);
	if (ret >= 0) 
		assignthread(ret,path);
	return ret;
}


int creat64(const char *path, mode_t mode)
{
	int ret;
	init();
	ret = realcreat64(path,mode);
	debug("creat64(\"%s\",%x) = %d\n",path,mode,ret);
	if (ret >= 0) 
		assignthread(ret,path);
	return ret;
}


#ifdef __sun
int openat64(int at,const char *fn, int flags, ...)
{
	int ret;

	init();
	if (flags & O_CREAT) {
		va_list val;
		mode_t mode;

		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
		ret = realopenat64(at,fn,flags,mode);
		debug("openat64(%d,\"%s\",%x,%x) = %d\n",at,fn,flags,mode,ret);
	} else {
		ret = realopenat64(at,fn,flags);
		debug("openat64(%d,\"%s\",%x) = %d\n",at,fn,flags,ret);
	}
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}


int openat(int at,const char *fn, int flags, ...)
{
	int ret;

	init();
	if (flags & O_CREAT) {
		va_list val;
		mode_t mode;

		va_start(val,flags);
		mode = va_arg(val,int);
		va_end(val);
		ret = realopenat(at,fn,flags,mode);
		debug("openat(%d,\"%s\",%x,%x) = %d\n",at,fn,flags,mode,ret);
	} else {
		ret = realopenat(at,fn,flags);
		debug("openat(%d,\"%s\",%x) = %d\n",at,fn,flags,ret);
	}
	if (!(flags & O_RDONLY) && (ret >= 0))
		assignthread(ret,fn);
	return ret;
}
#endif


FILE *fdopen(int fd, const char *mode)
{
	operator_t *op = Operators;
	init();
	FILE *ret = realfdopen(fd,mode);
	while (op && (op->fd != fd)) 
		op = op->next;
	if (op && op->fd == fd) {
		int err;
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		op->str = ret;
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		debug("fdopen(%d,\"%s\") = %p -> thread %d\n",fd,mode,ret,op->thread);
	} else
		debug("fdopen(%d,\"%s\") = %p\n",fd,mode,ret);
	return ret;
}


operator_t *add_op(op_t o, int fd, FILE *str)
{
	op_t *n;
	operator_t *op = Operators;
	int err;

	if (fd != -1) {
		while (op && (op->fd != fd)) 
			op = op->next;
	} else {
		while (op && (op->str != str)) 
			op = op->next;
	}
	if (op == 0)
		return 0;
	err = pthread_mutex_lock(&op->lock);
	assert(err == 0);
	if (op->unused) {
		n = op->unused;
		op->unused = n->next;
	} else {
		n = malloc(sizeof(op_t));
		assert(n);
	}
	*n = o;
	n->next = 0;
	if (op->head == 0) {
		op->head = op->tail = n;
	} else {
		op->tail->next = n;
		op->tail = n;
	}
	err = pthread_mutex_unlock(&op->lock);
	assert(err == 0);
	err = sem_post(&op->qlen);
	assert(err == 0);
	return op;
}


int fclose(FILE *str)
{
	int tid,ret;
	op_t o = {fc,0,0,0,0};
	operator_t *op;
	init();
	op = add_op(o,-1,str);
	if (op) {
		debug("fclose(%p [%d]) -> thread %d\n",str,op->fd,op->thread);
		return 0;
	}
	ret = realfclose(str);
	debug("realfclose(%p) = %d\n",str,ret);
	return ret;
}


int close(int fd)
{
	int tid,ret;
	op_t o = {cl,0,0,0,0};
	operator_t *op;
	init();
	op = add_op(o,fd,0);
	if (op) {
		debug("close(%d) -> thread %d\n",fd,op->thread);
		return 0;
	}
	ret = realclose(fd);
	debug("realclose(%d) = %d\n",fd,ret);
	return ret;
}


int write(int f, const void *b, size_t s)
{
	static const size_t minbufsize = 64 << 10;
	operator_t *op = Operators;
	init();
	while (op && (op->fd != f))
		op = op->next;
	if (op && op->fd == f) {
		int err;
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		if (op->wr_err != 0) {
			errno = op->wr_err;
			err = pthread_mutex_unlock(&op->lock);
			assert(err == 0);
			return -1;
		}
		if (MemPerThread) {
			while (((s < MemPerThread) && (s + op->mem > MemPerThread))
				|| ((s > MemPerThread) && (op->mem > 0))) {
				err = pthread_cond_wait(&op->oom,&op->lock);
				assert(err == 0);
			}
		}
		if (MemPerThread && (s > MemPerThread)) {
			int r;
			assert(op->head == 0);
			r = realwrite(f,b,s);
			++NumDirect;
			debug("nocopy realwrite(%d,%p,%d) = %d\n",f,b,s,r);
			s = r;
		} else if ((op->tail == 0)) {
			op_t *o;
			size_t bs = s < minbufsize ? minbufsize : s;
			assert(op->head == 0);
			if (op->unused) {
				o = op->unused;
				op->unused = o->next;
			} else {
				o = (op_t *) malloc(sizeof(op_t));
			}
			assert(o);
			o->op = wr;
			o->buf = malloc(bs);
			assert(o->buf);
			(void) memcpy(o->buf,b,s);
			o->s = s;
			o->bs = bs;
			o->next = 0;
			op->head = op->tail = o;
			op->mem += s;
			err = sem_post(&op->qlen);
			assert(err == 0);
			debug("write(%d,%p,%d) -> thread %d\n",f,b,s,op->thread);
		} else {
			op_t *o = op->tail;
			assert(o->next == 0);
			size_t os = o->s;
			size_t ns = os+s;
			if (ns > o->bs) {
				size_t nbs = (ns / minbufsize + 1) * minbufsize;
				o->buf = realloc(o->buf,nbs);
				o->bs = nbs;
			}
			op->mem += s;
			assert(o->buf);
			(void) memcpy((char *)o->buf+os,b,s);
			o->s = ns;
			debug("write(%d,%p,%d) -> thread %d\n",f,b,s,op->thread);
		}
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		return s;
	}
	debug("realwrite(%d,%p,%d)\n",f,b,s);
	return realwrite(f,b,s);
}


int fsync(int fd)
{
	operator_t *op;
	int ret = 0;
	op_t o = {fs,0,0,0,0};

	op = add_op(o,fd,0);
	if (op) {
		debug("fsync(%d) -> thread %d\n",fd,op->thread);
	} else {
		ret = realfsync(fd);
		debug("realfsync(%d) = %d\n",fd,ret);
	}
	return ret;
}


int fflush(FILE *str)
{
	operator_t *op;
	int ret = 0;
	op_t o = {ff,0,0,0,0};

	op = add_op(o,-1,str);
	if (op) {
		debug("fflush(%p) -> thread %d\n",str,op->thread);
	} else {
		ret = realfflush(str);
		debug("realfflush(%p) = %d\n",str,ret);
	}
	return ret;
}


static void *writer_thread(void *arg)
{
	operator_t *op = (operator_t *) arg;
	op_t *unused = 0;

	init();
	debug("thread %d: started\n",op->thread);
	for (;;) {
		op_t *o;
		int err;
		int fd;

		err = sem_wait(&op->qlen);
		assert(err == 0);
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		fd = op->fd;
		if (unused) {
			unused->next = op->unused;
			op->unused = unused;
		}
		assert(op->head);
		o = op->head;
		if (o == op->tail) {
			assert(o->next == 0);
			op->head = op->tail = 0;
		} else {
			op->head = o->next;
			assert(op->head);
		}
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		switch (o->op) {
		default: assert(0); break;

		case cl: {
				int clret;
				err = pthread_mutex_lock(&op->lock);
				assert(err == 0);
				clret = realclose(fd);
				if (clret == -1)
					error("error closing file %s: %s\n",op->fn,strerror(errno));
				debug("thread %d: realclose(%d) = %d\n",op->thread,fd,err);
				assert((op->head == 0)
					|| ((op->head->op == tr)
						&& (op->head->next == 0)));
				free(op->fn);
				op->fn = 0;
				op->fd = -1;
				err = pthread_mutex_unlock(&op->lock);
				assert(err == 0);
				err = sem_post(&Idle);
				assert(err == 0);
			 } break;

		case fc: {
				int ret;
				err = pthread_mutex_lock(&op->lock);
				assert(err == 0);
				ret = realfclose(op->str);
				if (ret != 0)
					error("error closing file %s: %s\n",op->fn,strerror(errno));
				debug("thread %d: realfclose(%p [= %d]) = %d\n",op->thread,op->str,fd,ret);
				free(op->fn);
				op->fd = -1;
				op->str = 0;
				op->fn = 0;
				err = pthread_mutex_unlock(&op->lock);
				assert(err == 0);
				err = sem_post(&Idle);
				assert(err == 0);
			 } break;

		case wr: 
			if (op->wr_err == 0) {
				int wrerr, num;
				num = realwrite(fd,o->buf,o->s);
				wrerr = errno;
				debug("thread %d: realwrite(%d,%p,%d) = %d\n",op->thread,fd,o->buf,o->s,err);
				free(o->buf);
				err = pthread_mutex_lock(&op->lock);
				assert(err == 0);
				op->mem -= o->s;
				err = pthread_cond_signal(&op->oom);
				assert(err == 0);
				if (num == -1) {
					int clerr, ret;
					op->wr_err = wrerr;
					error("error writing to file %s: %s\n",op->fn,strerror(wrerr));
					err = sem_post(&Idle);
					assert(err == 0);
				}
				err = pthread_mutex_unlock(&op->lock);
				assert(err == 0);
			} break;

		case fs:
			 if (-1 == realfsync(fd))
				error("thread %d: error syncing file %s: %s\n",op->thread,op->fn,strerror(errno));
			 break;

		case ff:
			 if (-1 == realfflush(op->str))
				error("thread %d: error flushing file %s: %s\n",op->thread,op->fn,strerror(errno));
			 break;

		case tr:
			debug("thread %d: exit\n",op->thread);
			pthread_exit(0);
			break;
		}
		unused = o;
	}
}


void exit(int e)
{
	init();
	debug("exit(%d)\n",e);
	operator_t *op = Operators;
	while (op) {
		int err;
		op_t *o = (op_t *)malloc(sizeof(op_t));
		o->op = tr;
		o->next = 0;
		err = pthread_mutex_lock(&op->lock);
		assert(err == 0);
		if (op->tail)
			op->tail->next = o;
		else
			op->head = o;
		op->tail = o;
		err = pthread_mutex_unlock(&op->lock);
		assert(err == 0);
		err = sem_post(&op->qlen);
		assert(err == 0);
		op = op->next;
	}
	op = Operators;
	while (op) {
		int err = pthread_join(op->thread,(void **)0);
		assert(err == 0);
		op = op->next;
	}
	if (Verbose) {
		log("NumThreads = %d\n",NumThreads);
		log("NumDirect = %d\n",NumDirect);
	}
	realexit(e);
#ifdef __GNUC__
	for (;;);	/* just to avoid a warning */
#endif
}


